/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.engvew.pruefung;

import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Iterator;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageCauser;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import de.bsvrz.vew.engvew.engstelle.Baustelle;
import de.bsvrz.vew.engvew.engstelle.Engstelle;
import de.bsvrz.vew.engvew.engstelle.ErgebnisEngpassPruefung;
import de.bsvrz.vew.engvew.engstelle.ErgebnisFsPruefung;
import de.bsvrz.vew.engvew.engstelle.ErgebnisStauPruefung;
import de.bsvrz.vew.engvew.engstelle.Situation;
import de.bsvrz.vew.engvew.engstelle.Unfall;
import de.bsvrz.vew.engvew.verwaltung.VerwaltungEngstellen;

/**
 * PruefEngstellen fhrt zyklischen Plausibilittsprfungen fr die 
 * von einer Instanz von <code>VerwaltungEngstellen</code> verwalteten 
 * Engstellen durch.
 *  
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.1 $ / $Date: 2008/10/31 13:14:13 $ / ($Author: Sans $)
 */
public class PruefEngstellen implements Runnable 
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private Debug _debug = Debug.getLogger();
 
  /**
   * Die Engstellenverwaltung
   */
  private VerwaltungEngstellen _engstVerw;
  
  /**
   * Zykluszeit in Sekunden
   */
  private int _zyklus;
  
  /**
   * Bestimmt die Datenverteiler-Verbindung
   * 
   * @return die Datenverteiler-Verbindung
   */
  public ClientDavInterface getConnection()
  {
    return _engstVerw.getConnection();
  }
   
  /**
   * Erzeugt ein neue Instanz vom Typ <code>PruefEngstellen</code>
   * und startet deren run()-Methode in einem neuen Thread
   * 
   * @param zyklus Zykluszeit
   * @param engstVerw Engestellenverwaltung
   */
  public PruefEngstellen(int zyklus, VerwaltungEngstellen engstVerw)
  {
    _zyklus = zyklus;
    _engstVerw = engstVerw;
    
    new Thread(this).start();
  }
  
  /*
   * (Kein Javadoc)
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    while(true)
    {  
      long start = getConnection().getTime();
           
      Iterator<Situation> it = _engstVerw.getVerwalteteObjekte().iterator();
      
      while(it.hasNext()) 
      {
        try
        {
          Engstelle engst = (Engstelle)it.next();
          stauPruefung(engst);
          engpassPruefung(engst);
          if (engst instanceof Unfall)
          {
            ablaufPruefung((Unfall)engst);
          }
          if (engst instanceof Baustelle)
          {
            gesperrteFahrstreifenPruefung((Baustelle)engst);
          }
        }  
        catch (ConcurrentModificationException e) // Engstellenmenge gendert
        {
          _debug.config("Engstellenmenge gendert: Abbruch laufende Prfung und Warten auf nchsten Zyklus");
          break;
        }
      }
      
      long ende = getConnection().getTime();
            
      _debug.config("Die Plausibilittsprfungen fr diesen Zyklus dauerten " +
          (ende - start) + "ms");
      
      if ((ende - start) < (_zyklus * 1000))
      { 
       
        _debug.config("Warte " +((_zyklus * 1000) - (ende - start)) +
            "ms auf nchsten Zyklus");
          getConnection().sleep((_zyklus * 1000) - (ende - start));
      }
    }
  }

  /**
   * Fhrt die Stauprfung fr eine Engstelle durch und versendet bei Bedarf eine entsprechende 
   * Betriebsmeldung
   * 
   * @param engst die Engstelle, die geprft werden soll
   */
  protected void stauPruefung(Engstelle engst)
  {
    try
    {
      Collection<ErgebnisStauPruefung> result = engst.pruefeStaus();
      for (ErgebnisStauPruefung ergebnis: result)
      {  
        String id = engst.getIdStauMeldung(ergebnis.stau);
        if (!engst.stauMeldungen.isMeldungAktiv(id))
        {  
        
          sendMessage("Ein Stau geht mehr als " + engst.getMaxLaengeStauUeberBaustellenEnde() +
                "m ("+ Math.round(ergebnis.ueberLaenge) +"m) ber Engstelle '" + engst.getObjekt().getName() + "' hinaus (Stau " + ergebnis.stau + 
                ")", engst.getObjekt());
                
          _debug.config("Betriebsmeldung Stau versendet fr " + engst + ", Stau " + ergebnis.stau);
          engst.setStauMeldungAktiv(ergebnis.stau);
        }
        else
        {  
          _debug.config("Betriebsmeldung Stau unterdrckt " + engst + ", Stau " + ergebnis.stau);
        }
      }  
    }
    catch (IllegalStateException e) // Engstelle nicht prfbar
    {
      _debug.info("Voraussetzungen fr Stau-Prfung nicht erfllt: " + e.getMessage());
    }
  }

  /**
   * Fhrt die Engpassprfung fr eine Engstelle durch und versendet bei Bedarf eine entsprechende 
   * Betriebsmeldung
   * 
   * @param engst die Engstelle, die geprft werden soll
   */
  protected void engpassPruefung(Engstelle engst)
  {
    try
    {
      Collection<ErgebnisEngpassPruefung> result = engst.pruefeEngpass();
      for (ErgebnisEngpassPruefung ergebnis: result)
      {  
        String id = engst.getIdKapazitaetsMeldung(ergebnis.mq);
        if (!engst.kapazitaetsMeldungen.isMeldungAktiv(id))
        {  
          sendMessage("Hinter der Engstelle '" + engst.getObjekt().getName() + "' (an MQ " + 
              ergebnis.mq + ") bersteigt die Verkehrstrke mit " + ergebnis.vs + " Fz/h die Restkapazitt von " +
              ergebnis.rk + " Fz/h um mehr als Faktor " + engst.getFaktorVerkehrsStaerkeStromAbwaerts(), 
              engst.getObjekt());
                
          _debug.config("Betriebsmeldung Engpass versendet fr " + engst + ", MQ " + ergebnis.mq);
          engst.kapazitaetsMeldungen.setMeldungAktiv(id);
        }
        else
        {
          _debug.config("Betriebsmeldung Engpass unterdrckt fr " + engst + ", MQ " + ergebnis.mq);
        }
      }   
    }
    catch (IllegalStateException e) // Engstelle nicht prfbar
    {
      _debug.info("Voraussetzungen fr Engpass-Prfung nicht erfllt: " + e.getMessage());
    }
  }
  
  /**
   * Fhrt die Ablaufprfung fr einen Unfall durch und versendet bei Bedarf eine entsprechende 
   * Betriebsmeldung
   * 
   * @param unfall der Unfall, der geprft werden soll
   */
  protected void ablaufPruefung(Unfall unfall)
  {
    try
    {
      long result = unfall.pruefeAblauf();
      if (result > 0)
      {  
        String id = unfall.getIdAblaufMeldung();
        if (!unfall.ablaufMeldungen.isMeldungAktiv(id))
        {  
          String wert = "";
          if (result > unfall.getWarteZeitMeldungNachRaeumungUnfall() / (60 * 1000))
          {  
            wert = "("+ result + " Minut(en)) ";
            if (result >= 24 * 60)
            {
              long d = result / (24 * 60);
              long m = result % (24 * 60);
              long  h = 0;
              if (m >= 60)
              {  
                h = m / 60;
                m = m % 60;
              }
              wert = "(" + d + " Tag(en)";
              if (h > 0)
              {
                wert += " " + h + " Stund(en)";
              }
              if (m > 0)
              {
                wert += " " + m + " Minut(en)";
              }
              wert += ") ";
            }
            
            if (result >60)
            {
              long  h = result / 60;
              long m = result % 60;
              wert = "(" + h + " Stund(en)";
              if (m > 0)
              {
                wert += " " + m + " Minut(en)";
              }
              wert += ") ";
            }
          }  
          
          sendMessage("Die (geschtzte) Dauer des Unfalls '" + unfall.getObjekt().getName() + 
              "' ist seit mehr als " + unfall.getWarteZeitMeldungNachRaeumungUnfall() / (60 * 1000) + 
              " Minuten " + wert + "abgelaufen", unfall.getObjekt());
                
          _debug.config("Betriebsmeldung Ablauf versendet fr " + unfall);
          unfall.ablaufMeldungen.setMeldungAktiv(id);
        }
        else
        {
          _debug.config("Betriebsmeldung Ablauf unterdrckt fr " + unfall);
        }
      }   
    }
    catch (IllegalStateException e) // Engstelle nicht prfbar
    {
      _debug.config("Voraussetzungen fr Zeit-Prfung nicht erfllt: " + e.getMessage());
    }
  }
  
  /**
   * Fhrt die Prfung der gesperrten Fahrstreifen fr eine Baustelle durch und versendet 
   * bei Bedarf eine entsprechende Betriebsmeldung
   * 
   * @param baust die Baustelle, die geprft werden soll
   */
  protected void gesperrteFahrstreifenPruefung(Baustelle baust)
  {
    try
    {
      Collection<ErgebnisFsPruefung> result = baust.pruefeGesperrteFahrstreifen();
      for (ErgebnisFsPruefung ergebnis: result)
      {  
        String id = baust.getIdFahrStreifenMeldung(ergebnis.fs);
        if (!baust.fahrStreifenMeldungen.isMeldungAktiv(id))
        {
          AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.fahrStreifen");
          String lage = ergebnis.fs.getConfigurationData(atg).getItem("Lage").asTextValue().getText();
          
          sendMessage("Auf dem gesperrten Fahrstreifen " + lage + " (Mestelle " + ergebnis.fs + ") der Baustelle '" + 
              baust.getObjekt().getName() + "' bersteigt die Verkehrstrke mit " + ergebnis.vs +" Fz/min den zulssigen Wert von " + 
              baust.getMaxVerkehrsStaerkeGesperrteFahrStreifen() + " Fz/min", baust.getObjekt());
                
          _debug.config("Betriebsmeldung gesperrter Fahrstreifen versendet fr " + baust + ", Fahrstreifen " + ergebnis.fs);
          baust.fahrStreifenMeldungen.setMeldungAktiv(id);

        }
        else
        {
          _debug.config("Betriebsmeldung gesperrter Fahrstreifen unterdrckt fr " + baust + ", Fahrstreifen " + ergebnis.fs);
        }
      }   
    }
    catch (IllegalStateException e) // Engstelle nicht prfbar
    {
      _debug.info("Voraussetzungen fr Prfung gesperrter Fahrstreifen nicht erfllt: " + e.getMessage());
    }
  }
  
  /**
   * Sendet eine Betriebsmeldung vom Meldungstyp APPLICATION_DOMAIN mit der 
   * Meldungsklasse WARNING
   *  
   * @param message Meldungstext
   * @param reference referenziertes Systemobjekt
   */
  protected void sendMessage(String message, SystemObject reference)
  {
    String id = "" + getConnection().getLocalApplicationObject().getId();
    sendMessage(id, MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, message, reference);
  }
  
  /**
   * Sendet eine Betriebsmeldung
   *  
   * @param type Meldungstyp
   * @param grade Meldungsklasse
   * @param message Meldungstext
   * @param reference referenziertes Systemobjekt
   */ 
  protected void sendMessage(String id, MessageType type, MessageGrade grade, String message, 
      SystemObject reference)
  {
    MessageSender.getInstance().sendMessage(id, type, null, grade, reference,  
        MessageState.MESSAGE, new MessageCauser(getConnection().getLocalUser(), "", ""), message);
  }
  
}
